#pragma once

#include <stdlib.h>
#include <new>
#include <limits>
#include <string.h>
#include <iomanip>

#define DYCV_TMP_MAX max
#undef max

#define DYCV_TMP_MIN min
#undef min

class TDAllocatorBase
{
public:

	TDAllocatorBase() {}
	virtual ~TDAllocatorBase() {}

	virtual void * Allocate(size_t size) = 0;

	virtual void Deallocate(void * p) = 0;
	
};

typedef TDAllocatorBase * PDAllocatorBase;

// align pointer

inline void * AlignedPointer(void * _p, size_t alsz)
{
	unsigned char * cp;
	unsigned char * alcp;
	void * ppt;

	cp = (unsigned char *)_p;
	alcp = (unsigned char *)(((size_t)(cp + alsz + sizeof(void *) - 1) / alsz) * alsz);
	ppt = (void *)(alcp - sizeof(void *));

	*(void * *)ppt = _p;

	return (void *)alcp;
}

inline void * UnAlignedPointer(void * _p)
{
	unsigned char * cp;
	void * ppt;
	cp = (unsigned char *)_p;

	cp = cp - sizeof(void *);
	ppt = *(void * *)cp;

	return ppt;
}

// new delete allocator

class TDAllocatorNew : public TDAllocatorBase
{
private:
	size_t alignment;
	size_t count;

public:

	TDAllocatorNew() : alignment(128), count(0) {}
	TDAllocatorNew(size_t _alignment) : alignment(_alignment), count(0) {}
	virtual ~TDAllocatorNew() {}

	void SetAlignment(size_t alignment);
	size_t GetAlignment();

	virtual void * Allocate(size_t size);
	virtual void Deallocate(void * p);

	bool IsEmpty();

};


// linked list allocator

struct TDAllocatorFitChunk;
typedef TDAllocatorFitChunk * PDAllocatorFitChunk;

struct TDAllocatorFitChunk
{
	PDAllocatorFitChunk next;
	PDAllocatorFitChunk prev;
	unsigned char * allocatedStart;
	unsigned char * freedStart;
	unsigned char * cunkEnd;
	size_t allocatedSize;

	void Init();

	void Remove();
	void InsertNext(PDAllocatorFitChunk nxt);
};

static const unsigned int DycNextFit = 0;
static const unsigned int DycFirstFit = 1;

class TDAllocatorFit : public TDAllocatorBase
{
private:
	size_t alignment;
	unsigned int fitType;
	bool antiFragmentation;
	size_t totalUsedSize;
	size_t totalAllocatedSize;

	unsigned char * memoryStart;
	unsigned char * memoryEnd;
	PDAllocatorFitChunk currentChunk;
	

	void MemInit();

public:

	TDAllocatorFit() : alignment(128), fitType(DycNextFit), antiFragmentation(true),
		memoryStart(NULL), memoryEnd(NULL) 
	{
		
	}

	TDAllocatorFit(size_t _alignment, unsigned int _fitType, bool _antiFragmentation,
			void * _memoryStart, void * _memoryEnd) : 
		alignment(_alignment), fitType(_fitType), antiFragmentation(_antiFragmentation),
		memoryStart((unsigned char *)_memoryStart), memoryEnd((unsigned char *)_memoryEnd) 
	{
		MemInit();
	}

	TDAllocatorFit(const TDAllocatorFit & a) : alignment(a.alignment), fitType(a.fitType), antiFragmentation(a.antiFragmentation),
		memoryStart(a.memoryStart), memoryEnd(a.memoryEnd) 
	{
		MemInit();
	}

	virtual ~TDAllocatorFit();

	void * GetMemoryStart();
	void * GetMemoryEnd();

	virtual void * Allocate(size_t size);
	virtual void Deallocate(void * p);

	size_t GetTotalUsedSize();
	size_t GetTotalAllocatedSize();

	size_t GetHeaderSize();

	bool IsEmpty();

};

// STL new delete allocator

template <class TDType> class TDSTLAllocator
{
public:
	typedef size_t size_type;
	typedef ptrdiff_t difference_type;
	typedef TDType * pointer;
	typedef const TDType * const_pointer;
	typedef TDType & reference;
	typedef const TDType & const_reference;
	typedef TDType value_type;

	template <class TDOther> struct rebind { typedef TDSTLAllocator<TDOther> other; };

public:

	pointer address(reference v) const
	{
		return (&v);
	}

	const_pointer address(const_reference v) const
	{
		return (&v);
	}

	TDSTLAllocator() throw() {}

	TDSTLAllocator(const TDSTLAllocator<TDType> & src) throw() {}

	template<class TDOther> 
		TDSTLAllocator(const TDSTLAllocator<TDOther> & src) throw() {}

	TDSTLAllocator& operator=( const TDSTLAllocator& ha )
	{
		return *this;
	}

	pointer allocate(size_type _Count, const_pointer _Hint = 0)
	{
		unsigned char * tp;
		tp = new unsigned char[_Count * sizeof(TDType)];

		return (pointer)tp;
	}

	void construct(pointer _Ptr, const TDType & _Value)
	{
		new((void*)_Ptr) TDType(_Value);
	}

	void deallocate(pointer _Ptr, size_type _Count)
	{
		delete [] (unsigned char *)_Ptr;
	}

	void destroy(pointer _Ptr)
	{
		_Ptr->~TDType();
	}

	size_type max_size() const throw()
	{
		return std::numeric_limits<size_t>::max() / sizeof(TDType);
	}

	//

	template<class TDTmpType> struct TDDestroyer
	{
		TDSTLAllocator<TDTmpType> a;

		TDDestroyer(TDSTLAllocator<TDTmpType> _a) : a(_a) {}

		TDDestroyer(const TDDestroyer & _d) : a(_d.a) {}

		void operator() ( TDTmpType * p ) 
		{
			a.destroy(p);
			a.deallocate(p, 1);
		}
	};

	
	//template<class TDTmpType> void CreateNewInstance(boost::shared_ptr<TDTmpType> & sp, const TDTmpType & c)
	//{
	//	typename rebind<TDTmpType>::other oa(*this);
	//
	//	TDTmpType * p;
	//	p = (TDTmpType *)oa.allocate(1, 0);
	//	new(p) TDTmpType(c);
	//
	//	TDDestroyer<TDTmpType> d(oa);
	//
	//	sp = boost::shared_ptr<TDTmpType>(p, d, oa);
	//	
	//}
};

template <typename T1> bool operator==( const TDSTLAllocator<T1>& lhs, const TDSTLAllocator<T1>& rhs) throw()
{
	return (bool)true;
}

template <typename T1> bool operator!=( const TDSTLAllocator<T1>& lhs, const TDSTLAllocator<T1>& rhs) throw()
{
	return (bool)false;
}

// STL allocator with TDAllocatorBase.

template <class TDType> class TDSTLAllocatorStatic : public TDSTLAllocator<TDType>
{ 

public:

	typedef size_t size_type;
	typedef ptrdiff_t difference_type;
	typedef TDType * pointer;
	typedef const TDType * const_pointer;
	typedef TDType & reference;
	typedef const TDType & const_reference;
	typedef TDType value_type;

	template <class TDOther> struct rebind { typedef TDSTLAllocatorStatic<TDOther> other; };

	static PDAllocatorBase & GetAllocator()
	{
		static PDAllocatorBase a;
		return a;
	}

	TDSTLAllocatorStatic() throw() {}

	TDSTLAllocatorStatic(TDAllocatorBase & a) throw()
	{
		GetAllocator() = &a;
	}

	TDSTLAllocatorStatic(const TDSTLAllocatorStatic<TDType> & src) throw() {}

	template<class TDOther> 
		TDSTLAllocatorStatic(const TDSTLAllocatorStatic<TDOther> & src) throw()
	{
		GetAllocator() = TDSTLAllocatorStatic<TDOther>::GetAllocator();
	}

	TDSTLAllocatorStatic& operator=( const TDSTLAllocatorStatic& ha )
	{
		return *this;
	}

	pointer allocate(size_type _Count, const_pointer _Hint = 0)
	{
		(void)_Hint;
		return (pointer)GetAllocator()->Allocate(_Count * sizeof(TDType));
	}

	void deallocate(pointer _Ptr, size_type _Count)
	{
		(void)_Count;
		GetAllocator()->Deallocate(_Ptr);
	}

	//
	template<class TDTmpType> struct TDDestroyer
	{
		TDSTLAllocatorStatic<TDTmpType> a;

		TDDestroyer(TDSTLAllocatorStatic<TDTmpType> _a) : a(_a) {}

		TDDestroyer(const TDDestroyer & _d) : a(_d.a) {}

		void operator() ( TDTmpType * p ) 
		{
			a.destroy(p);
			a.deallocate(p, 1);
		}
	};

	//template<class TDTmpType> void CreateNewInstance(boost::shared_ptr<TDTmpType> & sp, const TDTmpType & c)
	//{
	//	typename rebind<TDTmpType>::other oa(*this);

	//	TDTmpType * p;
	//	//p = (TDTmpType *)oa.allocate(1, 0);
	//	p = (TDTmpType *)GetAllocator()->Allocate(sizeof(TDTmpType));
	//	//new(p) TDTmpType(c);
	//	oa.construct(p, TDTmpType(c));

	//	TDDestroyer<TDTmpType> d(oa);

	//	sp = boost::shared_ptr<TDTmpType>(p, d, oa);
	//	
	//}

	//template<class TDTmpType, class TDTmpAllocator> void CreateNewInstance(
	//		TDTmpAllocator * ta, boost::shared_ptr<TDTmpType> & sp, const TDTmpType & c)
	//{
	//	typename rebind<TDTmpType>::other oa(*this);

	//	TDTmpType * p;
	//	p = (TDTmpType *)ta->Allocate(sizeof(TDTmpType));
	//	//new(p) TDTmpType(c);
	//	oa.construct(p, TDTmpType(c));

	//	TDDestroyer<TDTmpType> d(oa);

	//	sp = boost::shared_ptr<TDTmpType>(p, d, oa);
	//	
	//}

	//template<class TDTmpType> void CreateNewInstance(boost::shared_ptr<TDTmpType> & sp)
	//{
	//	typename rebind<TDTmpType>::other oa(*this);

	//	TDTmpType * p;
	//	//p = (TDTmpType *)oa.allocate(1, 0);
	//	p = (TDTmpType *)GetAllocator()->Allocate(sizeof(TDTmpType), (void *)&sp);
	//	new(p) TDTmpType(*this);
	//	//oa.construct(p, TDTmpType(c));

	//	TDDestroyer<TDTmpType> d(oa);

	//	sp = boost::shared_ptr<TDTmpType>(p, d, oa);
	//	
	//}

	//template<class TDTmpType, class TDTmpAllocator> void CreateNewInstance(
	//		TDTmpAllocator * ta, boost::shared_ptr<TDTmpType> & sp)
	//{
	//	typename rebind<TDTmpType>::other oa(*this);

	//	TDTmpType * p;
	//	p = (TDTmpType *)ta->Allocate(sizeof(TDTmpType));
	//	new(p) TDTmpType(*this);
	//	//oa.construct(p, TDTmpType(c));

	//	TDDestroyer<TDTmpType> d(oa);

	//	sp = boost::shared_ptr<TDTmpType>(p, d, oa);
	//	
	//}

};

template <typename T1> bool operator==( const TDSTLAllocatorStatic<T1>& lhs, const TDSTLAllocatorStatic<T1>& rhs) throw()
{
	return (bool)true;
}

template <typename T1> bool operator!=( const TDSTLAllocatorStatic<T1>& lhs, const TDSTLAllocatorStatic<T1>& rhs) throw()
{
	return (bool)false;
}

// Statefull STL allocator with TDAllocatorBase.

template <class TDType> class TDSTLAllocatorStatefull : public TDSTLAllocator<TDType>
{ 

public:

	typedef size_t size_type;
	typedef ptrdiff_t difference_type;
	typedef TDType * pointer;
	typedef const TDType * const_pointer;
	typedef TDType & reference;
	typedef const TDType & const_reference;
	typedef TDType value_type;

	template <class TDOther> struct rebind { typedef TDSTLAllocatorStatefull<TDOther> other; };

	PDAllocatorBase a;

	TDSTLAllocatorStatefull() throw() {}

	TDSTLAllocatorStatefull(TDAllocatorBase & a) throw() : a(&a) {}

	TDSTLAllocatorStatefull(const TDSTLAllocatorStatefull<TDType> & src) throw() : a(src.a) {}

	template<class TDOther> 
	TDSTLAllocatorStatefull(const TDSTLAllocatorStatefull<TDOther> & src) throw() : a(src.a)
	{
		
	}

	TDSTLAllocatorStatefull& operator=( const TDSTLAllocatorStatefull& ha )
	{
		a = ha.a;
		return *this;
	}

	pointer allocate(size_type _Count, const_pointer _Hint = 0)
	{
		(void)_Hint;
		return (pointer)a->Allocate(_Count * sizeof(TDType));
	}

	void deallocate(pointer _Ptr, size_type _Count)
	{
		(void)_Count;
		a->Deallocate(_Ptr);
	}

	//
	template<class TDTmpType> struct TDDestroyer
	{
		TDSTLAllocatorStatefull<TDTmpType> a;

		TDDestroyer(TDSTLAllocatorStatefull<TDTmpType> _a) : a(_a) {}

		TDDestroyer(const TDDestroyer & _d) : a(_d.a) {}

		void operator() ( TDTmpType * p ) 
		{
			a.destroy(p);
			a.deallocate(p, 1);
		}
	};

	//template<class TDTmpType> void CreateNewInstance(boost::shared_ptr<TDTmpType> & sp, const TDTmpType & c)
	//{
	//	typename rebind<TDTmpType>::other oa(*this);

	//	TDTmpType * p;
	//	//p = (TDTmpType *)oa.allocate(1, 0);
	//	p = (TDTmpType *)GetAllocator()->Allocate(sizeof(TDTmpType));
	//	//new(p) TDTmpType(c);
	//	oa.construct(p, TDTmpType(c));

	//	TDDestroyer<TDTmpType> d(oa);

	//	sp = boost::shared_ptr<TDTmpType>(p, d, oa);
	//	
	//}

	//template<class TDTmpType, class TDTmpAllocator> void CreateNewInstance(
	//		TDTmpAllocator * ta, boost::shared_ptr<TDTmpType> & sp, const TDTmpType & c)
	//{
	//	typename rebind<TDTmpType>::other oa(*this);

	//	TDTmpType * p;
	//	p = (TDTmpType *)ta->Allocate(sizeof(TDTmpType));
	//	//new(p) TDTmpType(c);
	//	oa.construct(p, TDTmpType(c));

	//	TDDestroyer<TDTmpType> d(oa);

	//	sp = boost::shared_ptr<TDTmpType>(p, d, oa);
	//	
	//}

	//template<class TDTmpType> void CreateNewInstance(boost::shared_ptr<TDTmpType> & sp)
	//{
	//	typename rebind<TDTmpType>::other oa(*this);

	//	TDTmpType * p;
	//	//p = (TDTmpType *)oa.allocate(1, 0);
	//	p = (TDTmpType *)GetAllocator()->Allocate(sizeof(TDTmpType), (void *)&sp);
	//	new(p) TDTmpType(*this);
	//	//oa.construct(p, TDTmpType(c));

	//	TDDestroyer<TDTmpType> d(oa);

	//	sp = boost::shared_ptr<TDTmpType>(p, d, oa);
	//	
	//}

	//template<class TDTmpType, class TDTmpAllocator> void CreateNewInstance(
	//		TDTmpAllocator * ta, boost::shared_ptr<TDTmpType> & sp)
	//{
	//	typename rebind<TDTmpType>::other oa(*this);

	//	TDTmpType * p;
	//	p = (TDTmpType *)ta->Allocate(sizeof(TDTmpType));
	//	new(p) TDTmpType(*this);
	//	//oa.construct(p, TDTmpType(c));

	//	TDDestroyer<TDTmpType> d(oa);

	//	sp = boost::shared_ptr<TDTmpType>(p, d, oa);
	//	
	//}

};

template <typename T1> bool operator==( const TDSTLAllocatorStatefull<T1>& lhs, const TDSTLAllocatorStatefull<T1>& rhs) throw()
{
	return lhs.a == rhs.a;
}

template <typename T1> bool operator!=( const TDSTLAllocatorStatefull<T1>& lhs, const TDSTLAllocatorStatefull<T1>& rhs) throw()
{
	return lhs.a != rhs.a;
}

#define max DYCV_TMP_MAX
#define min DYCV_TMP_MIN

